package kripos.gateway;

import java.util.concurrent.ConcurrentHashMap;
import jade.content.ContentElement;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.OntologyException;
import jade.content.onto.UngroundedException;
import jade.content.onto.basic.Action;
import jade.content.onto.basic.Result;
import jade.core.AID;
import jade.core.ContainerID;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.JADEAgentManagement.WhereIsAgentAction;
import jade.domain.mobility.MobilityOntology;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.proto.SubscriptionInitiator;

/**
 * Subscribes to changes in registered AdminAgents by the AMS Service.
 * Any changes are recorded in a ConcurrentHashMap 
 * that was first initialized by setup(). 
 * 
 * @author oysteine
 * @version 1.0
 */
public class AdminSubscribeBehaviour extends SubscriptionInitiator {

  private static final long serialVersionUID = 988748612411732543L;
  private GWAgent myAgent;
  private ConcurrentHashMap<AID, String> myMap;

  /**
   * Constructor
   * 
   * @param a The agent this instance belongs to
   * @param msg The message used by the behaviour to initiate subscription
   * @param map The hashmap for recording changes.
   */
  public AdminSubscribeBehaviour(GWAgent a, ACLMessage msg, ConcurrentHashMap<AID, String> map) {
    super(a, msg);
    myAgent = a;
    myMap = map;
  }

  /**
   * Handles incoming inform messages that contains changes in registered AdminAgents.
   * The AMS is queried for the Location of any new AdminAgents.
   * Removed AdminAgents are delete from the hashmap.
   */
  @Override
  protected void handleInform(ACLMessage inform) {
    try {
      DFAgentDescription[] dfds = 
        DFService.decodeNotification(inform.getContent());
      for(int i=0; i<dfds.length;i++){
        AID aid = dfds[i].getName();

        if(myMap.containsKey(aid)){
          /* There seems to be a problem with the AMS, messages containing 
           * removal information are sent without reason. This functionality disabled.
               myMap.remove(aid); 
           */       
        }
        else{
          WhereIsAgentAction wiaa = new WhereIsAgentAction();
          wiaa.setAgentIdentifier(aid);
          AID myAMS = myAgent.getAMS();
          Action act = new Action(myAMS, wiaa);

          ACLMessage qMsg = new ACLMessage(ACLMessage.REQUEST);
          String convID = myAgent.getConvList().registerConversation();
          qMsg.setConversationId(convID);

          qMsg.setLanguage(new SLCodec(0).getName());
          qMsg.setOntology(MobilityOntology.getInstance().getName());
          try {
            myAgent.getContentManager().fillContent(qMsg, act);
            qMsg.addReceiver(act.getActor());
            //register conversation with agent to get correct reception
            myAgent.send(qMsg);

            MessageTemplate mt = MessageTemplate.MatchConversationId(convID);
            ACLMessage resp = myAgent.blockingReceive(mt, 1000000);
            myAgent.getConvList().deregisterConversation(convID);
            ContentElement ce = myAgent.getContentManager().extractContent(resp);
            Result result = (Result) ce;
            ContainerID cid = (ContainerID)result.getValue();
            myMap.put(aid, cid.getName());

          } catch (UngroundedException e) {
            e.printStackTrace();
          } catch (CodecException e) {
            e.printStackTrace();
          } catch (OntologyException e) {
            e.printStackTrace();
          }
        }
      }
    }
    catch (FIPAException fe) {
      fe.printStackTrace();
    }
  }

}//class