package kripos.gateway;

import java.math.BigInteger;
import java.util.GregorianCalendar;
import java.util.Iterator;
import java.util.Set;
import jade.content.lang.Codec.CodecException;
import jade.content.lang.sl.SLCodec;
import jade.content.onto.OntologyException;
import jade.content.onto.basic.Action;
import jade.core.AID;
import jade.core.behaviours.SimpleBehaviour;
import jade.domain.mobility.MobilityOntology;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import kripos.ontology.*;

/**
 * Launched by GWAgent when a CommandPackage containing instructions to commence a trace is received
 * 
 * @author oysteine
 * @version 1.0
 */
public class LaunchTraceBehave extends SimpleBehaviour {
  private static final long serialVersionUID = 4144187902498096653L;
  private CommandPackage command;
  private GWAgent myGWAgent;
  private boolean done;

  /**
   * Creates an instance of this behaviour
   * 
   * @param a the GWAgent owning this instance
   * @param the CommandPackage containing information about what to do
   */
  public LaunchTraceBehave(GWAgent a, CommandPackage cp) {
    super(a);
    command = cp;
    myGWAgent = a;
  }

  /* (non-Javadoc)
   * @see jade.core.behaviours.Behaviour#action()
   */
  @Override
  public void action() {
    //parse package
    command.getReqTime();
    command.getTarget();
    command.getType();

    //construct and fill CreateCase, Case and Session
    CreateCase cc = new CreateCase();
    OntAddress target;
    if(isIP()){
      IP ip = new IP();
      ip.setHasName(command.getTarget());
      target = ip;
    }
    else{
      DNSname dns = new DNSname();
      dns.setHasName(command.getTarget());
      target = dns;
    }
    Case c = new Case();
    c.setCaseID((int)Math.round((Math.random())*100000000));
    c.setCaseName("testcase"); //todo get from GUI

    Session s = new Session();
    s.setCaseID(c.getCaseID());

    s.setHasBaseAddress(target);
    c.setCaseSessions(s);
    cc.setHasCase(c);

    OntDate date = new OntDate();
    long time = GregorianCalendar.getInstance().getTimeInMillis();
    date.setTime(time);
    cc.setHasDate(date);
    date.setTime(time+10000);
    c.setCaseStartDate(date);

    Action act = new Action(getRandomAdminAgent(),cc);
    String convID = sendRequest(act);
    myGWAgent.addCommand(convID, command);
    MessageTemplate mt = MessageTemplate.MatchConversationId(convID);
    myAgent.blockingReceive(mt);
  }

  /**
   * 
   * @return true if the String represents and IP address and not DNS name
   */
  private boolean isIP(){
    String toParse = command.getTarget().replace('.', '0');
    try{
      new BigInteger(toParse);
      return true;
    }
    catch(NumberFormatException nfe){
      return false;
    }
  }

  /**
   * Get a random AdminAgent from the hashmap of AdminAgents
   * 
   * @return the AID of the selected AdminAgent
   */
  private AID getRandomAdminAgent(){
    int size = myGWAgent.adminAgents.size();
    int random = (int)Math.round((Math.random())*size);
    Set<AID> tempSet = myGWAgent.adminAgents.keySet();
    Iterator<AID> it = tempSet.iterator();
    for(int i=0;i<random-1;i++){
      it.next();
    }
    return it.next();
  }

  /**
   * Utility method for sending ACLMessages 
   * <p>
   * Registers the conversationID in the convList 
   * to avoid loosing any reply 
   * 
   * @param action the action that is to be wrapped in an ACLMessage 
   * @return the conversationID for the created ACLMessage
   */
  private String sendRequest(Action action){
    try {
      ACLMessage qMsg = new ACLMessage(ACLMessage.REQUEST);
      String convID = myGWAgent.getConvList().registerConversation();
      qMsg.setConversationId(convID);
      qMsg.setLanguage(new SLCodec(0).getName());
      qMsg.setOntology(MobilityOntology.getInstance().getName());
      myAgent.getContentManager().fillContent(qMsg, action);
      qMsg.addReceiver(action.getActor());
      //register conversation with agent to get correct reception of reply
      myAgent.send(qMsg);
      return convID;
    } 
    catch (CodecException e) {
      e.printStackTrace();
      return null;
    } 
    catch (OntologyException e) {
      e.printStackTrace();
      return null;
    }
  }

  /* (non-Javadoc)
   * @see jade.core.behaviours.Behaviour#done()
   */
  @Override
  public boolean done() {
    return done;
  }

}//class