package kripos.geo;

import jade.content.ContentElement;
import jade.content.lang.Codec.CodecException;
import jade.content.onto.OntologyException;
import jade.content.onto.UngroundedException;
import jade.content.onto.basic.Action;
import jade.core.behaviours.OneShotBehaviour;
import jade.lang.acl.ACLMessage;
import kripos.ontology.Case;
import kripos.ontology.Report;
import kripos.ontology.Session;

/**
 * Called when a WorkerAgent receives a request to 
 * report the status of its session.
 * <p>
 * The content of the report differs depending on the 
 * type of work the WorkerAgent is doing.
 * <p>
 * Generate a reply to its SessionAgent
 * 
 * @author oysteine
 * @version 1.0
 */
public class WorkerReportBehaviour extends OneShotBehaviour {
   private static final long serialVersionUID = 3252684517173485891L;
   private WorkerAgent myWorkerAgent;
   private ACLMessage myMsg;

   /**
    * Creates an instance of this behaviour
    *  
    * @param a the WorkerAgent that owns this instance
    * @param m the ACLMessage this behaviour will work on
    */
   public WorkerReportBehaviour(WorkerAgent a, ACLMessage m) {
      super(a);
      myWorkerAgent = a;
      myMsg = m;
   }

   /**
    * Check if the Session to report on is the one the owning WorkerAgent is working on.
    * If it is, report back to the SessionAgent of the owning WorkerAgent.
    * 
    * @see jade.core.behaviours.Behaviour#action()
    */
   @Override
   public void action() {		
      try {
         //identify Case to be reported
         ContentElement ce = myAgent.getContentManager().extractContent(myMsg);
         Report report = (Report)((Action)ce).getAction();
         Case ctbr = report.getCaseToBeReported();
         Session sess = ctbr.getCaseSessions();
         if (myWorkerAgent.getSessionID() == sess.getSessionID()){
            report();
         }
         else{
            ACLMessage reply = myMsg.createReply();
            reply.setPerformative(ACLMessage.NOT_UNDERSTOOD);
            myAgent.send(reply);
         }

      } catch (UngroundedException e) {
         e.printStackTrace();
      } catch (CodecException e) {
         e.printStackTrace();
      } catch (OntologyException e) {
         e.printStackTrace();
      }
   }//action()

   /**
    * Does the actual work of generating the report, and send the reply.
    */
   private void report(){
      ACLMessage reply = myMsg.createReply();

      if (myWorkerAgent.getType().equalsIgnoreCase("CRAWL")){
         //not implemented yet
         //report number of files secured, names?, types?
         //report number of found remaining files
         reply.setPerformative(ACLMessage.NOT_UNDERSTOOD);
      }
      else if (myWorkerAgent.getType().startsWith("TRACE")){
         //not implemented yet
         //anything to report here? depends on implementation!
         reply.setPerformative(ACLMessage.NOT_UNDERSTOOD);
      }
      else{
         reply.setPerformative(ACLMessage.NOT_UNDERSTOOD);
      }
      myAgent.send(reply);
   }

}//class