package kripos.geo.openmap;


import java.awt.Color;
import java.awt.Point;
import java.awt.Shape;
import java.awt.geom.Area;
import java.awt.geom.FlatteningPathIterator;
import java.awt.geom.GeneralPath;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.locks.ReentrantLock;

import kripos.math.circle.Circle;
import kripos.math.circle.Intersection;
import kripos.math.circle.Intersector;
import kripos.math.servlet.GetMap;

import com.bbn.openmap.LatLonPoint;
import com.bbn.openmap.Layer;
import com.bbn.openmap.MapBean;
import com.bbn.openmap.event.LayerStatusEvent;
import com.bbn.openmap.event.LayerStatusListener;
import com.bbn.openmap.image.AcmeGifFormatter;
import com.bbn.openmap.layer.OMGraphicHandlerLayer;
import com.bbn.openmap.layer.location.LocationHandler;
import com.bbn.openmap.layer.location.LocationLayer;
import com.bbn.openmap.layer.location.csv.CSVLocationHandler;
import com.bbn.openmap.layer.shape.ShapeLayer;
import com.bbn.openmap.omGraphics.OMCircle;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMGraphicList;
import com.bbn.openmap.omGraphics.OMPoint;
import com.bbn.openmap.omGraphics.OMPoly;
import com.bbn.openmap.proj.Length;
import com.bbn.openmap.proj.Mercator;
import com.bbn.openmap.proj.Proj;
import com.bbn.openmap.proj.Projection;

/**
 * Draw circles on a map.
 * @author oysteine
 */
public class MapDrawer {

  private static final int MAX_CACHE = 100;
  private static final Object cacheMutex = new Object();
  private static final Map<Long,CacheEntry> mapCache =
    new HashMap<Long,CacheEntry>();

  private static final List<Long> cacheLru = new LinkedList<Long>();

  private static class CacheEntry {
    private final Long cacheId;
    private final ReentrantLock lock = new ReentrantLock();
    /** Set <i>after</i> all other data have been set. */
    private MapBean mapBean = null;
    private LatLonPoint pointUpperLeft;
    private LatLonPoint pointLowerRight;
    private LatLonPoint pointCenter;
    private float scale;
    private float origScale;

    private double area;
    private LatLonPoint centroid = null;

    private LayerListener listener;
    private LocationHandler locationHandler;

    private CacheEntry(Long cacheId) {
      this.cacheId = cacheId;
    }
  }

  private static CacheEntry acquireMap(Long cacheId) {
    CacheEntry ce;
    synchronized (cacheMutex) {
      ce = mapCache.get(cacheId);
      if (ce == null) {
        ce = unsafeNewCacheEntry(cacheId);
      }

      cacheLru.remove(cacheId);
      cacheLru.add(cacheId);
    }

    ce.lock.lock();
    return ce;
  }

  private static void releaseMap(CacheEntry ce) {
    synchronized (cacheMutex) {
      ce.lock.unlock();
      cacheLru.add(ce.cacheId);
    }
  }


  private static CacheEntry unsafeNewCacheEntry(Long cacheId) {
    assert Thread.holdsLock(cacheMutex);

    CacheEntry ce = new CacheEntry(cacheId);
    mapCache.put(cacheId, ce);
    if (mapCache.size() > MAX_CACHE) {
      mapCache.remove(cacheLru.remove(0));
    }
    return ce;
  }



  /**
   * 
   * @param cacheId use cached drawing instead of creating using <code>circles</code>,
   *          if the cache holds anything
   * @param circles
   * @return informat about area and centroid that will be drawn on map
   */
  static public IntersectionInfo getIntersectionInfo(
      Long cacheId, List<Circle> circles) {

    CacheEntry ce = acquireMap(cacheId);
    try {
      if (ce.mapBean == null) {
        createMap(ce, circles);
      }

      if (ce.centroid == null) {
        return new IntersectionInfo(
            ce.area,
            ce.mapBean.getWidth(),
            ce.mapBean.getHeight());
      } else {
        return new IntersectionInfo(
            ce.area,
            ce.centroid.getLatitude(),
            ce.centroid.getLongitude(),
            ce.mapBean.getWidth(),
            ce.mapBean.getHeight());
      }

    } finally {
      releaseMap(ce);
    }
  }



  private static final Proj dummyProj =
    new Mercator(new LatLonPoint(0.123456, -0.123456), 1.42E7f, 640, 480);


  /**
   * @param cacheId use cached drawing instead of creating using <code>circles</code>,
   *          if the cache holds anything
   * @param circles
   * @param zoom 1.0 is with polygon extremes on the borders,
   *          smaller values are larger portion of the world,
   *          greater values are zoomed inside polygon
   * @param latOffset north offset of quasi center of polygon in decimal degrees 
   * @param longOffset east offset of quasi center of polygon in decimal degrees 
   * @return a map with <code>circles</code> painted on it
   */
  static public byte[] getMap(Long cacheId, List<Circle> circles,
      float zoom, double latOffset, double longOffset) {

    CacheEntry ce = acquireMap(cacheId);
    try {
      if (ce.mapBean == null) {
        createMap(ce, circles);
      }



      LatLonPoint customCenter = new LatLonPoint(
          ce.pointCenter.getLatitude() + latOffset,
          ce.pointCenter.getLongitude() + longOffset);
      float customScale = ce.scale / zoom;
      if (customScale > ce.origScale) {
        customScale = ce.origScale;
      }

      // disable city names if showing too much of world at once
      ce.locationHandler.setShowNames(customScale < 2.5E7);

      // update map
      Proj proj = new Mercator(customCenter, customScale,
          ce.mapBean.getWidth(), ce.mapBean.getHeight());

      // if we use only default parameters the first time, then for some reason:
      // - traceLayer is not drawn
      // - repaints hang in LayerListener
      // so we use a dummy projection first
      ce.listener.resetCompletion(1);
      ce.mapBean.setProjection(dummyProj);
      ce.listener.waitForCompletion();

      ce.listener.resetCompletion(1);
      ce.mapBean.setProjection(proj);
      ce.listener.waitForCompletion();

      // create image
      AcmeGifFormatter gifFormatter = new AcmeGifFormatter();
      byte[] image = gifFormatter.getImageFromMapBean(ce.mapBean);
      return image;

    } finally {
      releaseMap(ce);
    }
  }




  private static void createMap(CacheEntry ce, List<Circle> circles) {
    String shapeFile = GetMap.getDataPath() + "/shape/dcwpo-browse.shp";
    String spatialFile = GetMap.getDataPath() + "/shape/dcwpo-browse.ssx";
    String locationFile = GetMap.getDataPath() + "/cities.csv";

    ce.listener = new LayerListener();

    MapBean mapBean = new MapBean();
    mapBean.setSize(1024, 768);
    MapBean.suppressCopyright = true; // suppress after first use
    // political borders
    ShapeLayer backgroundLayer = new ShapeLayer();
    ce.listener.addLayer(backgroundLayer, 1);
    Properties backgroundProps = new Properties();
    backgroundProps.put("prettyName", "Political Solid");
    backgroundProps.put("lineColor", "000000");
    backgroundProps.put("fillColor", "BDDE83");
    backgroundProps.put("shapeFile", shapeFile);
    backgroundProps.put("spatialIndex", spatialFile);
    backgroundLayer.setProperties(backgroundProps);
    assert mapBean.getComponentCount() == 0 : mapBean.getComponentCount();
    mapBean.add(backgroundLayer, 0);

    OMGraphicList traceItems = new OMGraphicList(2);
    OMGraphicList tracePolygon = new OMGraphicList(1);
    OMGraphicList traceCircles = new OMGraphicList(circles.size());

    // inaccurate intersections used for initial scale calculation
    List<Intersection> inters2D = Intersector.getIntersections(circles);
    inters2D = Intersector.getMorePoints(inters2D, 0.2);

    setBorders(ce, inters2D, traceItems, mapBean);
    setBorderInfo(ce, traceItems);

    setPolygon2D(inters2D, tracePolygon);
    setCircles(circles, traceCircles);

    OMGraphicHandlerLayer traceLayer = new OMGraphicHandlerLayer();
    ce.listener.addLayer(traceLayer, 1);
    traceItems.add(tracePolygon);
    traceItems.add(traceCircles);
    traceLayer.setList(traceItems);
    mapBean.add(traceLayer, 0);
    ce.listener.waitForCompletion();

    List<Point2D> screenPath = getWindingPath(ce, traceCircles, mapBean);
    Point centroidXY = getScreenCentroid(screenPath);
    if (centroidXY == null) {
      ce.centroid = null;
    } else {
      ce.centroid = mapBean.getProjection().inverse(centroidXY);
    }
    List<LatLonPoint> approximatedPath = new LinkedList<LatLonPoint>();
    for(Point2D next : screenPath){
      Point temp = new Point((int)next.getX(), (int)next.getY());
      approximatedPath.add(mapBean.getProjection().inverse(temp));
    }
    ce.area = getApproximatedPolygonArea(approximatedPath);

    traceItems = new OMGraphicList(2);
    tracePolygon = new OMGraphicList(1);
    traceCircles = new OMGraphicList(circles.size());

    setBorderInfo(ce, traceItems);
    setApproximatedPolygon(approximatedPath, tracePolygon);
    setCircles(circles, traceCircles);

    LocationLayer locations = new LocationLayer();
    ce.listener.addLayer(locations, 1);
    ce.locationHandler = new CSVLocationHandler();
    Properties lhProps = new Properties();
    lhProps.put("locationFile", locationFile);
    lhProps.put("csvFileHasHeader", "false");
    lhProps.put("showNames", "true");
    //lhProps.put("nameColor", "008C54");
    lhProps.put("nameColor", "000000");
    lhProps.put("showLocations", "true");
    lhProps.put("locationColor", "FF0000");

    lhProps.put("name.lineColor", "FF008C54");
    lhProps.put("location.lineColor", "FFFF0000");
    lhProps.put("location.fillColor", "FFaaaaaa");
    lhProps.put("location.pointRadius", "2");
    lhProps.put("location.pointOval", "true");

    lhProps.put("nameIndex", "0");
    lhProps.put("latIndex", "2");
    lhProps.put("lonIndex", "3");
    ce.locationHandler.setProperties(lhProps);

    //locations.setProperties(lhProps);
    locations.setLocationHandlers(new LocationHandler[] { ce.locationHandler });

    Proj proj = new Mercator(ce.pointCenter, ce.scale,
        mapBean.getWidth(), mapBean.getHeight());
    ce.listener.resetCompletion(1);
    traceItems.add(tracePolygon);
    traceItems.add(traceCircles);
    traceLayer.setList(traceItems);
    mapBean.add(locations, 0);
    mapBean.setProjection(proj);
    ce.listener.waitForCompletion();  

    ce.mapBean = mapBean;
  }

  private static double getApproximatedPolygonArea(List<LatLonPoint> sortedPath){
    if (sortedPath.size() < 3) {
      return 0.0;
    }

    List<LatLonPoint> tmp = new LinkedList<LatLonPoint>(sortedPath);
    LatLonPoint base = tmp.remove(0);

    int triangleCount = 0;
    double interiorAngleSum = 0.0;
    LatLonPoint prev = tmp.remove(0);
    for (LatLonPoint next : tmp) {
      LatLonPoint A = base;
      LatLonPoint B = prev;
      LatLonPoint C = next;
      double a = B.distance(C);
      double b = C.distance(A);
      double c = A.distance(B);

      if (a >= Math.ulp(a) && b >= Math.ulp(b) && c >= Math.ulp(c)) {
        double angleA = Math.acos(
            (Math.cos(a) - Math.cos(b) * Math.cos(c)) /
            (Math.sin(b) * Math.sin(c)));
        double angleB = Math.acos(
            (Math.cos(b) - Math.cos(c) * Math.cos(a)) /
            (Math.sin(c) * Math.sin(a)));
        double angleC = Math.acos(
            (Math.cos(c) - Math.cos(b) * Math.cos(a)) /
            (Math.sin(a) * Math.sin(b)));

        interiorAngleSum += angleA + angleB + angleC;
        triangleCount++;
      } else {
        // 1 distance is 0, the other two are equal, OR
        // all distances are 0
        //assert a < Math.ulp(a) : A + "," + B + "," + C;
        //assert b < Math.ulp(b) : A + "," + B + "," + C;
        //assert c < Math.ulp(c) : A + "," + B + "," + C;
      }

      prev = next;
    }

    // approx, should use rad at current geodetic latitude
    double earthRad = (6378.135 + 6356.750) / 2.0; 
    double area =
      Math.pow(earthRad, 2) *
      (interiorAngleSum - Math.PI * triangleCount); 
    return area;
  }

  private static double getScreenPolygonArea(List<Point2D> sortedPath){
    if (sortedPath.size() == 0) {
      return 0.0;
    }

    double area = 0.0;

    List<Point2D> tmp = new LinkedList<Point2D>(sortedPath);
    Point2D prev = tmp.remove(0);
    tmp.add(prev);
    for (Point2D next : tmp) {
      double x1 = prev.getX();
      double y1 = prev.getY();
      double x2 = next.getX();
      double y2 = next.getY();
      area += x1*y2 - x2*y1;
      prev = next;
    }
    area *= 0.5;
    return area;
  }

  private static Point getScreenCentroid(List<Point2D> points)
  {
    if (points.size() < 2) {
      return null;
    }

    double area = getScreenPolygonArea(points);

    double x = 0.0;
    double y = 0.0;

    List<Point2D> tmp = new LinkedList<Point2D>(points);
    Point2D prev = tmp.remove(0);
    tmp.add(prev);
    for (Point2D next : tmp) {
      double x1 = prev.getX();
      double y1 = prev.getY();
      double x2 = next.getX();
      double y2 = next.getY();

      double determinant = x1*y2 - x2*y1;
      x += (x1+x2) * determinant;
      y += (y1+y2) * determinant;

      prev = next;
    }

    x /= (6 * area);
    y /= (6 * area);

    return new Point((int)x,(int)y);
  }

  private static List<Point2D> getWindingPath(
      CacheEntry ce, OMGraphicList circles,
      MapBean mapBean) {

    List<Point2D> ret = new LinkedList<Point2D>();
    if (circles.size() == 0) {
      return ret;
    }

    Proj proj = new Mercator(ce.pointCenter, ce.scale,
        mapBean.getWidth(), mapBean.getHeight());
    ce.listener.resetCompletion(1);
    mapBean.setProjection(proj);
    ce.listener.waitForCompletion();

    Iterator iter = circles.iterator();
    OMCircle circle = (OMCircle)iter.next();

    Shape s = circle.getShape();
    assert s != null;
    Area polygon = new Area(s);

    while (iter.hasNext()) {
      circle = (OMCircle)iter.next();
      s = circle.getShape();
      assert s != null;
      Area area = new Area(s);
      polygon.intersect(area);
    }

    PathIterator pathIter = polygon.getPathIterator(null);
    GeneralPath genPath = new GeneralPath();
    int theType;
    float[] theData = new float[6];

    while(!pathIter.isDone()){
      theType = pathIter.currentSegment(theData);

      switch(theType){
      case PathIterator.SEG_MOVETO :
        genPath.moveTo(theData[0], theData[1]);
        break;
      case PathIterator.SEG_LINETO :
        genPath.lineTo(theData[0], theData[1]);
        break;
      case PathIterator.SEG_QUADTO :
        genPath.quadTo(theData[0], theData[1], theData[2], theData[3]);
        break;
      case PathIterator.SEG_CUBICTO :
        genPath.curveTo(theData[0], theData[1], theData[2], theData[3], theData[4], theData[5]);
        break;
      case PathIterator.SEG_CLOSE :
        genPath.closePath();
        break;
      }//end switch

      pathIter.next();
    }

    PathIterator pathIter2 = genPath.getPathIterator(null);
    FlatteningPathIterator fpi = new FlatteningPathIterator(pathIter2, 0.25);
    double[] coords = new double[6];
    while (!fpi.isDone()) {
      fpi.currentSegment(coords);
      ret.add(new Point2D.Double(coords[0],  coords[1]));
      fpi.next();
    }

    return ret;
  }


  private static void setApproximatedPolygon(List<LatLonPoint> approximatePath,
      OMGraphicList traceItems) {

    if (approximatePath.size() > 0) {
      float[] lats_lons = new float[2*(approximatePath.size()+1)];
      int i = 0;
      for (LatLonPoint point : approximatePath) {
        lats_lons[i++] = (float)point.getLatitude();
        lats_lons[i++] = (float)point.getLongitude();
      }
      LatLonPoint firstPoint = approximatePath.get(0);
      lats_lons[i++] = (float)firstPoint.getLatitude();
      lats_lons[i++] = (float)firstPoint.getLongitude();
      OMPoly omPolygon = new OMAlphaPoly(lats_lons, OMGraphic.DECIMAL_DEGREES,
          OMGraphic.LINETYPE_STRAIGHT, 0.1f);
      omPolygon.setLinePaint(Color.black);
      omPolygon.setFillPaint(Color.red);
      traceItems.add(omPolygon);
    }
  }


  private static void setBorders(CacheEntry ce, List<Intersection> inters,
      OMGraphicList traceItems, MapBean mapBean) {
    double northMost = Double.MIN_VALUE;
    double southMost = Double.MAX_VALUE;
    double westMost = Double.MAX_VALUE;
    double eastMost = Double.MIN_VALUE;

    if (inters.size() == 0) {
      northMost = 90.0;
      southMost = -90.0;
      westMost = -180.0;
      eastMost = 180.0;
    } else {
      for (Intersection inter : inters) {
        double longitude = inter.point.getY();
        double latitude = inter.point.getX();

        if (latitude > northMost) {
          northMost = latitude;
        }
        if (latitude < southMost) {
          southMost = latitude;
        }
        if (longitude < westMost) {
          westMost = longitude;
        }
        if (longitude > eastMost) {
          eastMost = longitude;
        }
      }
    }

    ce.pointUpperLeft = new LatLonPoint(northMost, westMost);
    ce.pointLowerRight = new LatLonPoint(southMost, eastMost);

    Projection prevProj = mapBean.getProjection();
    Point pUL = prevProj.forward(ce.pointUpperLeft);
    Point pLR = prevProj.forward(ce.pointLowerRight);
    ce.origScale = prevProj.getScale();
    ce.scale = 2f * prevProj.getScale(
        ce.pointUpperLeft, ce.pointLowerRight, pUL, pLR);

    ce.pointCenter = new LatLonPoint(
        northMost - (northMost - southMost) / 2,
        westMost  + (eastMost - westMost) / 2);
  }

  private static void setBorderInfo(
      CacheEntry ce, OMGraphicList traceItems) {

    if (ce.centroid != null) {
      OMPoint p = new OMPoint(
          ce.centroid.getLatitude(),
          ce.centroid.getLongitude(),
          5);
      p.setFillPaint(Color.pink);
      traceItems.add(p);
    }
  }


  private static void setPolygon2D(List<Intersection> inters,
      OMGraphicList traceItems) {

    if (inters.size() > 0) {
      float[] lats_lons = new float[2*(inters.size()+1)];
      int i = 0;
      for (Intersection inter : inters) {
        lats_lons[i++] = (float)inter.point.getX();
        lats_lons[i++] = (float)inter.point.getY();
      }
      Intersection firstInter = inters.get(0);
      lats_lons[i++] = (float)firstInter.point.getX();
      lats_lons[i++] = (float)firstInter.point.getY();
      OMPoly omPolygon = new OMAlphaPoly(lats_lons, OMGraphic.DECIMAL_DEGREES,
          OMGraphic.LINETYPE_STRAIGHT, 0.4f);
      omPolygon.setLinePaint(Color.black);
      omPolygon.setFillPaint(Color.red);
      traceItems.add(omPolygon);
    }
  }


  private static void setCircles(List<Circle> circles,
      OMGraphicList traceItems) {
    for (Circle circle : circles) {
      Point2D p = circle.origo;
      LatLonPoint center = new LatLonPoint(p.getX(), p.getY());
      OMCircle omCircle = new OMAlphaCircle(
          center, (float)circle.rad, Length.DECIMAL_DEGREE, 0, 0.2f);
      omCircle.setLinePaint(Color.black);
      omCircle.setFillPaint(Color.black);
      traceItems.add(omCircle);
    }
  }


  private static class LayerListener implements LayerStatusListener {

    private final Object completionMutex = new Object();
    private final Map<Layer,Integer> completionMap =
      new HashMap<Layer,Integer>();


    private void addLayer(Layer layer, int remainingCount) {
      synchronized (completionMutex) {
        layer.addLayerStatusListener(this);
        completionMap.put(layer, remainingCount);
      }
    }

    public void updateLayerStatus(LayerStatusEvent evt) {
      synchronized (completionMutex) {
        switch (evt.getStatus()) {
        case LayerStatusEvent.DISTRESS:
          break;
        case LayerStatusEvent.FINISH_WORKING:
          completionMap.put(evt.getLayer(),
              completionMap.get(evt.getLayer()) - 1);
          completionMutex.notifyAll();
          break;
        case LayerStatusEvent.START_WORKING:
          break;
        case LayerStatusEvent.STATUS_UPDATE:
          break;
        }
      }
    }


    private void resetCompletion(int remainingCount) {
      synchronized (completionMutex) {
        for (Layer layer : completionMap.keySet()) {
          completionMap.put(layer, remainingCount);
        }
      }
    }

    private void waitForCompletion() {
      synchronized (completionMutex) {
        while (true) {
          boolean allDone = true;
          for (Integer remaining: completionMap.values()) {
            if (remaining > 0) {
              allDone = false;
              break;
            }
          }
          if (allDone) {
            break;
          }
          try {
            long startTime = System.currentTimeMillis();
            completionMutex.wait(5000);
            long endTime = System.currentTimeMillis();
            if (endTime - startTime > 4500) {
              break;
            }
          } catch (InterruptedException e) {
            throw new RuntimeException(e);
          }
        }
      }
    }
  }

}//class