package kripos.geo.ping;
//TODO support ping4 and ping6 simultaneously
/*
 * Some of the functionality of this class is adapted from RTTOmeter. 
 * (Code originally released under GPLv2)
 * The RTTOmeter application is available from: http://idmaps.eecs.umich.edu/rttometer/
 * The GPLv2 is available at: http://www.gnu.org/copyleft/gpl.html
 **/

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Arrays;

/**
 * 
 * @author oysteine
 * @version 1.0
 *
 */
public class MinRTT {
  private float CI = 0;
  private float CII = 0;
  private	float CIII = 0;
  private ArrayList<PingItem> probeList = new ArrayList<PingItem>();
  private float minimumRTT = 99999.0f;
  private float maximumRTT = 0;
  private float epsilon = MINEPS;//automatically changed based on minimumRTT

  /* default values of epsilon */
  public final static float UNINETTEPS = 0.7f; //single AS, most paths <<20ms
  public final static float MINEPS = 2;  //<=50ms
  public final static float MEDEPS = 4;  //50-150ms
  public final static float BIGEPS = 6;  //>150ms

  private float CI_threshold = 0.8f; //confidence in CI
  private final int EPS_MIN_PROBES = 100; //magic number. only for measuring between landmarks
  private boolean estimate_CI_confidence;

  private int listCounter = 0; //used to resume calculations if the probeList is extended
  private float accCi = 0; //cumulative
  private float accCii = 0; //cumulative
  private float accCiii = 0; //cumulative
  private int accN = 0; //cumulative

  private int numProbes; //Updated if successive runtimes
  private int rePingCnt = 0;
  //used by rePing()
  private String previousAddress;
  private int previousNumPings;
  private double previousPingDistance;

  /**
   * 
   */
  public MinRTT(){
  }

  /**
   *  
   * @param address
   * @param numPings
   * @param pingDistance
   * @throws HostUnreachableException
   */
  public ArrayList ping(String address, int numPings, double pingDistance, 
      boolean reset) throws HostUnreachableException{
    if(reset){
      reset();
    }

    previousAddress = address;
    previousNumPings = numPings;
    previousPingDistance = pingDistance;

    int lastSequence = 0;
    listCounter = 0;

    try {
      Runtime runtime = Runtime.getRuntime();
      String command = "ping -i"+pingDistance+" -c"+numPings+" " + address;
      Process p = runtime.exec(command);
      numProbes += numPings;
      InputStream ip = p.getInputStream();
      InputStreamReader isr = new InputStreamReader(ip, Charset.forName("ISO-8859-1"));
      BufferedReader br = new BufferedReader(isr);

      String temp = null;
      while((temp = br.readLine()) !=null){
        //handle unreachable host
        if(temp.contains("0 received, 100% packet loss")){
          for(int i=0;i<numPings;i++){
            probeList.add( new PingItem(PingItem.LOST, lastSequence+i+1,-999999));
            listCounter++;
          }
          throw new HostUnreachableException("100% packet loss");
        }

        //sort out filtered replies
        if(temp.contains("Packet filtered")){
          int cutSub = temp.indexOf("icmp_seq=");
          String subTemp = temp.substring(cutSub);
          int cutSeq = subTemp.indexOf('=');
          int cutSeqEnd = subTemp.indexOf("Packet filtered");
          String seqString = subTemp.substring(cutSeq+1,cutSeqEnd-1);
          int seq = Integer.parseInt(seqString);

          int seqDelta = seq - lastSequence;
          if(seqDelta > 1){
            for(int i=0; i<seqDelta;i++){
              PingItem lostPi = new PingItem(PingItem.LOST, lastSequence+i,-999999);
              probeList.add(lostPi);
              listCounter++;
            }
          }

          lastSequence = seq;
          PingItem pi = new PingItem(PingItem.FILTERED, seq, -999999);
          probeList.add(pi);
          listCounter++;
          throw new HostUnreachableException("Packet filtering on path to host");
        }

        //normal pingitem, if not 64 bytes it is not a regular 
        //ICMP_ECHO_REPLY packet and we ignore it
        if(temp.contains("64 bytes from")){
          int cutPoint = temp.indexOf("=");
          String cutTemp = temp.substring(cutPoint);
          int seqEndPoint = cutTemp.indexOf("ttl");
          String seqString = cutTemp.substring(1,seqEndPoint-1);
          int seq = Integer.parseInt(seqString);

          //adds missing probes as they are not listed by ping's output
          int seqDelta = seq - lastSequence;
          if(seqDelta > 1){
            for(int i=0; i<seqDelta;i++){
              PingItem lostPi = new PingItem(PingItem.LOST, lastSequence+i,-999999);
              probeList.add(lostPi);
              listCounter++;
            }
          }

          lastSequence = seq;
          int timePoint = temp.lastIndexOf("=");
          int endPoint = temp.lastIndexOf("ms");
          String time = temp.substring(timePoint+1, endPoint-1);
          float rtt = Float.parseFloat(time);

          PingItem pi = new PingItem(PingItem.NORMAL_RTT, seq, rtt);
          probeList.add(pi);
          listCounter++;
        }

        if(temp.contains("rtt min/avg/max/mdev = ")){

          //common variables
          int cutPoint = temp.indexOf("=");
          String cutTemp = temp.substring(cutPoint+2);
          int firstSlashPoint = cutTemp.indexOf("/");

          //maxRTT  //not needed
          int maxTempEndPointLong = cutTemp.indexOf("ms");
          String maxTempLong = cutTemp.substring(0, maxTempEndPointLong);
          int maxTempEndPoint = maxTempLong.lastIndexOf("/");
          String maxTemp = maxTempLong.substring(0, maxTempEndPoint);
          int maxEndPoint = maxTemp.lastIndexOf("/");
          int maxPoint = maxTemp.indexOf("/");
          String max = maxTemp.substring(maxPoint+1, maxEndPoint);
          float maxRtt = Float.parseFloat(max);
          if (maximumRTT < maxRtt){
            maximumRTT = maxRtt;
          }

          //minRTT
          String min = cutTemp.substring(0, firstSlashPoint); 
          float minRtt = Float.parseFloat(min);
          if(minimumRTT > minRtt){
            minimumRTT = minRtt;
          }
        }//if

      }//while
    }//try
    catch (IOException e) {
      throw new HostUnreachableException(e.getMessage(),e.getCause());
    }
    return probeList;
  }

  /**
   * 
   * 
   * Method adapted from RTTOmeter application
   */
  public int computeCRegions(boolean estimateConfidence){
    estimate_CI_confidence = estimateConfidence;
    float ci = 0;
    float cii = 0;
    float ciii = 0;
    int n = 0; //total number of phaseplot points (not probes)
    float val = 0;
    float w1 = 0;
    float w2 = 0;
    float w = 0;
    int loopStart = 0;

    if(minimumRTT <20){
      epsilon = UNINETTEPS;
    }
    else if(minimumRTT <50){
      epsilon = MINEPS;
    }
    else if (minimumRTT <150){
      epsilon = MEDEPS;
    }
    else{
      epsilon = BIGEPS;
    }

    if(probeList.size() != listCounter){
      loopStart = (probeList.size() - listCounter)-1; //include the last probe from the previous run  
    }

    for(int i=loopStart;i<probeList.size()-1;i++) {//we do not want to compare the last pingitem to null 
      PingItem piCurrent = probeList.get(i);
      PingItem piNext = probeList.get(i+1);

      //Classify the point in the phase-plot
      if(piCurrent.getType() == PingItem.NORMAL_RTT && piNext.getType() == PingItem.NORMAL_RTT ){
        n++;

        /* Point in CI, CII, or CIII */
        if( (piCurrent.getRTT() - minimumRTT) <= epsilon){
          w1 = 1.0f;
        }
        else {
          val = ((int)((piCurrent.getRTT() - minimumRTT)/epsilon) + 1);
          w1 = 1.0f/val;
        }

        if( (piNext.getRTT() - minimumRTT) <= epsilon){
          w2 = 1.0f;
        }
        else {
          val = ((int)((piNext.getRTT() - minimumRTT)/epsilon) + 1);
          w2 = 1.0f/val;
        }

        w = w1 * w2;
      }
      else {
        /* Point in CII or CIII */
        if(piCurrent.getType() == PingItem.LOST) {
          w1 = 0.0f;
        } else {
          val = ((int)((piCurrent.getRTT() - minimumRTT)/epsilon) + 1);
          w1 = 1.0f/val;
        }

        if(piNext.getType() == PingItem.LOST) {
          w2 = 0.0f;
        } else {
          val = ((int)((piNext.getRTT() - minimumRTT)/epsilon) + 1);
          w2 = 1.0f/val;
        }

        w = w1 * w2;
      }

      ci += w;
      if(w1 == 1 || w2 == 1){
        /* This point in CII */
        cii += (1-w);
      }
      else{
        /* Definitely CIII */
        ciii += (1-w);
      }
    }
    accCi=+ci;
    accCii=+cii;
    accCiii=+ciii;
    accN=+n;

    if(accN > 0) {
      CI = accCi/accN; 
      CII =  accCii/accN; 
      CIII =  accCiii/accN; 
    }
    if(estimate_CI_confidence && CI< CI_threshold && rePingCnt <= 2){
      rePing();
    }
    return n;

  }//compute

  /**
   * 
   *
   */
  private void rePing() {//reuses original input. TODO:estimate better values. wait() if c2 high?
    rePingCnt++;
    try {
      ping(previousAddress, previousNumPings, previousPingDistance, false);
      computeCRegions(estimate_CI_confidence);
    } catch (HostUnreachableException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  private void reset(){
    accCi = 0;
    accCii = 0;
    accCiii = 0;
    accN = 0;
    rePingCnt = 0;
    numProbes = 0;
    CI = 0;
    CII = 0;
    CIII = 0;
    probeList.clear();
    minimumRTT = 99999.0f;
    maximumRTT = 0;
  }

  /**
   * Estimates epsilon from the ping measurements 
   * 
   * Method adapted from RTTOmeter application
   */
  public float estimateEpsilon(){
    int n = 0;
    float eps = -1.0f;
    float[] rtts = new float[numProbes];

    /* make array of rtts from probelist */
    for(int i=0;i<probeList.size();i++) { 
      PingItem piCurrent = probeList.get(i);
      if(piCurrent.getType() == PingItem.NORMAL_RTT ){
        rtts[n++] = piCurrent.getRTT();
      }
    }

    if (n >= (0.8 * EPS_MIN_PROBES)) { /* At least 80% of the probes are successful */
      Arrays.sort(rtts);
      ModeNode mn = new ModeNode();
      float mode_all = mn.mode(rtts, n);

      /* Estimate epsilon */
      eps = 2*(mode_all - rtts[0]); //or use minimumRTT, same value

      /* Make sure eps >=0, mode may be very close to min, 
       * and due to precision the subtraction might give negative result
       */
      if(eps < 0){
        eps = 0.0f;
      }
    }
    return eps;
  }

  /*
   * Getter methods
   */

  public float getCI(){
    return CI;
  }

  public float getCII(){
    return CII;
  }

  public float getCIII(){
    return CIII;
  }

  public float getMaxRTT(){
    return maximumRTT;
  }

  public float getMinRTT(){
    return minimumRTT;
  }

}//class
