package kripos.math.servlet;


import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.Locale;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import kripos.geo.openmap.IntersectionInfo;
import kripos.geo.openmap.MapDrawer;

/**
 * Display a page with the results of a trace.
 * The trace is performed if not cached already.
 * 
 * @author oysteine
 */
public class GetMap extends HttpServlet {
  private static final long serialVersionUID = 7631170050282471534L;

  private static String BASE_PATH = null;


  /**
   * @return path for OpenMap data. Does <i>not</i> end with separator.
   */
  public static String getDataPath() {
    String path = BASE_PATH;
    if (path == null) {
      path = System.getProperty("user.dir") + File.separator + "data";
    }
    return path;
  }



  @Override
  public void doGet(HttpServletRequest request, HttpServletResponse response)
  throws IOException, ServletException
  {
    ServletContext sc = getServletContext();
    if (BASE_PATH == null) {
      BASE_PATH = sc.getRealPath("") + File.separator + "data";
    }

    String ip = request.getParameter("ip");
    String cacheId = request.getParameter("cacheId");

    String strZoom = request.getParameter("zoom");
    String strLongOffset = request.getParameter("longOffset");
    String strLatOffset = request.getParameter("latOffset");
    float zoom = 1.0f;
    if (strZoom != null) {
      zoom = Float.parseFloat(strZoom);
    }
    double longOffset = 0.0;
    if (strLongOffset != null) {
      longOffset = Double.parseDouble(strLongOffset);
    }
    double latOffset = 0.0;
    if (strLatOffset != null) {
      latOffset = Double.parseDouble(strLatOffset);
    }



    String action = request.getParameter("action");
    boolean forceTrace = false;
    String retrieval;
    if ("Force trace".equals(action)) {
      forceTrace = true;
      retrieval = "Forced new trace";
    } else if ("Get trace".equals(action)) {
      if (TraceCache.hasTraceId(ip)) {
        retrieval = "Using cached trace";
      } else {
        retrieval = "Performed new trace";
      }
    } else if ("Zoom in".equals(action)) {
      retrieval = "Refreshed map of previous trace";
    } else if ("Zoom out".equals(action)) {
      retrieval = "Refreshed map of previous trace";
    } else if ("Move up".equals(action)) {
      retrieval = "Refreshed map of previous trace";
    } else if ("Move down".equals(action)) {
      retrieval = "Refreshed map of previous trace";
    } else if ("Move left".equals(action)) {
      retrieval = "Refreshed map of previous trace";
    } else if ("Move right".equals(action)) {
      retrieval = "Refreshed map of previous trace";
    } else {
      sc.log("Unknown action '" + action + "'");
      response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      return;
    }

    if (ip == null) {
      sc.log("Destination not specified!");
      response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      return;
    }

    // Perform trace or retrieve cached trace
    if (cacheId == null) {
      cacheId = TraceCache.getTraceId(ip, forceTrace);
    }
    TraceCacheEntry ce = TraceCache.getCachedTrace(cacheId, ip);
    if (ce == null) {
      sc.log("Could not find cache id '" + cacheId + "' for address '" + ip + "'.");
      response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
      return;
    }

    IntersectionInfo intersectionInfo = MapDrawer.getIntersectionInfo(ce.cacheId, ce.circles);
    //try { Thread.sleep(500); } catch (Exception e) {}


    response.setContentType("text/html");
    PrintWriter out = response.getWriter();
    out.println("<html>");
    out.println("<head>");
    out.println("<title>Trace of " + ip + "</title>");
    out.println("</head>");
    out.println("<body>");

    out.println("<h3>Trace info</h3>");
    out.println("Retrieval: " + retrieval + "<br>");
    out.println("IP traced: " + ip + "<br>");
    out.println("Trace started: " + ce.start + "<br>");
    out.println("Trace time: " + ce.runTime + "<br>");
    out.printf("<i>The trace information is cached for %d hours " +
        "after last usage</i><br>%n",
        (TraceCache.CACHE_LIFE_TIME / 3600000));
    out.println("<h3>Trace results</h3>");
    out.printf(Locale.US,
        "Estimated confidence region: %,.0f km<sup>2</sup><br>%n",
        intersectionInfo.polygonArea);
    out.println("Estimated location: ");
    if (!intersectionInfo.centroidAvailable) {
      out.println("(not available)");
    } else {
      out.println("longitude=" + degToString(intersectionInfo.centroidLongitude) +
          ", latitude=" + degToString(intersectionInfo.centroidLatitude) +
      "<br>");
    }
    out.println("<table><tr>");
    out.println("<td><form action=\"Map\" method=POST>");
    out.print("<input type=hidden name=ip value=\"" + ip + "\">");
    out.print("<input type=hidden name=cacheId value=\"" + cacheId + "\">");
    out.print("<input type=hidden name=zoom value=\"" + (zoom * 2) + "\">");
    out.print("<input type=hidden name=longOffset value=\"" + (longOffset) + "\">");
    out.print("<input type=hidden name=latOffset value=\"" + (latOffset) + "\">");
    out.println("<input type=submit name=action value=\"Zoom in\">");
    out.println("</form></td>");
    out.println("<td><form action=\"Map\" method=POST>");
    out.print("<input type=hidden name=ip value=\"" + ip + "\">");
    out.print("<input type=hidden name=cacheId value=\"" + cacheId + "\">");
    out.print("<input type=hidden name=zoom value=\"" + (zoom / 2) + "\">");
    out.print("<input type=hidden name=longOffset value=\"" + (longOffset) + "\">");
    out.print("<input type=hidden name=latOffset value=\"" + (latOffset) + "\">");
    out.println("<input type=submit name=action value=\"Zoom out\">");
    out.println("</form></td>");
    out.println("<td><form action=\"Map\" method=POST>");
    out.print("<input type=hidden name=ip value=\"" + ip + "\">");
    out.print("<input type=hidden name=cacheId value=\"" + cacheId + "\">");
    out.print("<input type=hidden name=zoom value=\"" + (zoom) + "\">");
    out.print("<input type=hidden name=longOffset value=\"" + (longOffset) + "\">");
    out.print("<input type=hidden name=latOffset value=\"" + (latOffset+2) + "\">");
    out.println("<input type=submit name=action value=\"Move up\">");
    out.println("</form></td>");
    out.println("<td><form action=\"Map\" method=POST>");
    out.print("<input type=hidden name=ip value=\"" + ip + "\">");
    out.print("<input type=hidden name=cacheId value=\"" + cacheId + "\">");
    out.print("<input type=hidden name=zoom value=\"" + (zoom) + "\">");
    out.print("<input type=hidden name=longOffset value=\"" + (longOffset) + "\">");
    out.print("<input type=hidden name=latOffset value=\"" + (latOffset-2) + "\">");
    out.println("<input type=submit name=action value=\"Move down\">");
    out.println("</form></td>");
    out.println("<td><form action=\"Map\" method=POST>");
    out.print("<input type=hidden name=ip value=\"" + ip + "\">");
    out.print("<input type=hidden name=cacheId value=\"" + cacheId + "\">");
    out.print("<input type=hidden name=zoom value=\"" + (zoom) + "\">");
    out.print("<input type=hidden name=longOffset value=\"" + (longOffset-2) + "\">");
    out.print("<input type=hidden name=latOffset value=\"" + (latOffset) + "\">");
    out.println("<input type=submit name=action value=\"Move left\">");
    out.println("</form></td>");
    out.println("<td><form action=\"Map\" method=POST>");
    out.print("<input type=hidden name=ip value=\"" + ip + "\">");
    out.print("<input type=hidden name=cacheId value=\"" + cacheId + "\">");
    out.print("<input type=hidden name=zoom value=\"" + (zoom) + "\">");
    out.print("<input type=hidden name=longOffset value=\"" + (longOffset+2) + "\">");
    out.print("<input type=hidden name=latOffset value=\"" + (latOffset) + "\">");
    out.println("<input type=submit name=action value=\"Move right\">");
    out.println("</form></td>");
    out.println("</tr></table>");
    out.println("<img src=\"MapImage?" +
        "ip=" + ip +
        "&cacheId=" + cacheId +
        "&zoom=" + zoom +
        "&longOffset=" + longOffset +
        "&latOffset=" + latOffset +
        "\"" +
        " border=1" +
        " width=" + intersectionInfo.imageWidth +
        " height=" + intersectionInfo.imageHeight +
    " alt=\"OpenMap(tm) image\"><br>");

    out.println("</body>");
    out.println("</html>");
  }

  @Override
  public void doPost(HttpServletRequest request, HttpServletResponse response)
  throws IOException, ServletException
  {
    doGet(request, response);
  }


  private String degToString(double degrees) {
    int deg = (int)degrees;
    double minutes = (degrees - deg) * 60;

    ByteArrayOutputStream tmp = new ByteArrayOutputStream();
    PrintStream stream = new PrintStream(tmp);
    stream.printf(Locale.US, "%d&deg; %.2f'", deg, minutes);
    return tmp.toString();
  }

}
