package kripos.math.servlet;

import jade.util.BasicProperties;
import jade.wrapper.ControllerException;
import jade.wrapper.StaleProxyException;
import jade.wrapper.gateway.JadeGateway;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;

import kripos.gateway.CommandPackage;
import kripos.math.circle.Circle;
import kripos.ontology.GeoLocation;
import kripos.ontology.TraceResult;

/**
 * Perform IP tracing and add the results to cache.
 * 
 * @author oysteine
 */
public class TraceCache {


  /** Max life time of a cache entry in milliseconds. */
  static final long CACHE_LIFE_TIME = 24 * 3600 * 1000L;

  static private final Random rand = new Random();


  private final static Object cacheMutex = new Object();

  /** Mapping from cacheId. */
  private final static Map<Long,TraceCacheEntry> idMap =
    new HashMap<Long,TraceCacheEntry>();

  /**
   * Mapping from IP to list of cacheId for all results
   * performed for within cache life time.
   * Most recent trace is stored first in the list. */
  private final static Map<String,List<Long>> destinationMap =
    new HashMap<String,List<Long>>();


  /**
   * Check if <code>destination</code> is cached.
   * 
   * @param ip
   * @return <code>true</code> iff <code>destination</code> can
   *  be retrieved with {@link #getTraceId(String, boolean)}
   *  without performing a trace
   * 
   */
  static boolean hasTraceId(String destination) {
    synchronized (cacheMutex) {
      List<Long> tmp = destinationMap.get(destination);
      if (tmp != null) {
        Long id = tmp.get(0);
        TraceCacheEntry ce = idMap.get(id);
        ce.lastAccess = new Date(); // to make sure it's still available when retrieving it a little later
        return true;
      } else {
        return false;
      }
    }
  }    
  /**
   * Perform a trace or return most recent previous trace of <code>ip</code>,
   * if one has been done.
   * 
   * @param ip a valid IP address
   * @param forceNew if <code>true</code> a new trace will always be performed,
   *          even if <code>ip</code> already exists in the cache
   * @return a cache id that can be used with {@link #getCachedTrace(String,String)}
   */
  static String getTraceId(String destination, boolean forceNew) {
    cleanCache();

    Long id = null;
    if (!forceNew) {
      synchronized (cacheMutex) {
        List<Long> tmp = destinationMap.get(destination);
        if (tmp != null) {
          id = tmp.get(0);
        }
      }
    }

    if (id == null) {
      // NOTE: it is possible that two traces are performed with the same IP
      //       at the same time
      TraceCacheEntry ce = performTrace(destination);
      insertToCache(ce);
      id = ce.cacheId;
    }

    return id.toString();
  }


  /**
   * Retrieve the trace specified by <code>cacheId</code> from cache.
   * 
   * @param cacheId     an id retrieved by {@link #getTrace(String, boolean)} recently
   * @param destination used to assert that the <code>cacheId</code>
   *          entry was performed with this destination.
   *          If it was not, the request is treated as if <code>cacheId</code>
   *          was not found
   * @return <code>null</code> if <code>cacheId</code> is not malformed,
   *          it does not exist, or
   *          it does not match <code>destination</code>
   *          
   */
  static TraceCacheEntry getCachedTrace(String cacheId, String destination) {
    cleanCache();

    Long id = null;
    try {
      id = Long.parseLong(cacheId);
    } catch (NumberFormatException e) {
      return null;
    }

    TraceCacheEntry ce;
    synchronized (cacheMutex) {
      ce = idMap.get(id);
    }

    if (ce == null || !ce.destination.equals(destination)) {
      return null;
    }

    return ce;
  }


  static private void insertToCache(TraceCacheEntry ce) {
    synchronized (cacheMutex) {
      long id = -1;
      synchronized (rand) {
        while (id < 0 || idMap.containsKey(id)) {
          id = rand.nextLong();
        }
      }
      ce.cacheId = id;

      idMap.put(ce.cacheId, ce);

      List<Long> tmp = destinationMap.get(ce.destination);
      if (tmp == null) {
        tmp = new LinkedList<Long>();
        destinationMap.put(ce.destination, tmp);
      }
      tmp.add(0, ce.cacheId);
    }
  }

  /**
   * Remove entries from cache if they are older than
   * {@linkplain #CACHE_LIFE_TIME} ms.
   */
  static private void cleanCache() {
    synchronized (cacheMutex) {
      long curTime = System.currentTimeMillis();
      for (Map.Entry<String,List<Long>> entry : destinationMap.entrySet()) {
        List<Long> idList = entry.getValue();
        Iterator<Long> iter = idList.iterator();
        while (iter.hasNext()) {
          Long id = iter.next();
          TraceCacheEntry ce = idMap.get(id);
          if (curTime - ce.lastAccess.getTime() > CACHE_LIFE_TIME) {
            idMap.remove(id);
            iter.remove();
          }
        }
        if (idList.size() == 0) {
          destinationMap.remove(entry.getKey());
        }
      }
    }
  }



  /**
   * @param destination a unique destination identifier
   * @return resulting circles from a trace
   */
  static private TraceCacheEntry performTrace(String destination) {
    long startTime = System.currentTimeMillis();
    List<Circle> circles = TRACELIBRARY_getTrace(destination);
    long endTime = System.currentTimeMillis();
    return new TraceCacheEntry(
        destination, circles,
        new Date(startTime), endTime - startTime);
  }


  /**
   * Interfaces with the multi-agent system to perfrom a trace
   * 
   * @param the IP address to trace
   * @return the geographical constraint circles
   */
  static private List<Circle> TRACELIBRARY_getTrace(String destination) {
    double KMperDegree = 1/111.15;
    List<Circle> circles = new LinkedList<Circle>();

    BasicProperties prop = new BasicProperties();
    prop.setProperty("platform-id",	"futurum01.item.ntnu.no:1099/JADE");//hard coded for now :(
    prop.setBooleanProperty("main", true);
    prop.setProperty("mainURL","http://futurum01.item.ntnu.no:7778/acc");//hard coded for now :(
    JadeGateway.init("kripos.gateway.GWAgent", prop);
    CommandPackage cp = new CommandPackage();
    cp.setTarget(destination);
    cp.setType("TRACE-CBG");

    try {

      JadeGateway.execute(cp);
    } catch (StaleProxyException e) {
      e.printStackTrace();
    } catch (ControllerException e) {
      e.printStackTrace();
    } catch (InterruptedException e) {
      e.printStackTrace();
    }

    ArrayList<TraceResult> traceRes = cp.getCBGResults();
    for(TraceResult tr : traceRes){
      GeoLocation geoLoc = tr.getHasGeoLocation();
      double lat = geoLoc.getLocationLatitude();
      double lon = geoLoc.getLocationLongitude();
      double radius = tr.getTraceResultData() * KMperDegree;
      circles.add(new Circle(lat, lon, radius));
    }

    return circles;
  }

}//class